/* packet-snmp.c
 * Routines for SNMP (simple network management protocol)
 * D.Jorand (c) 1998
 *
 * $Id: packet-snmp.c,v 1.28 2000/04/13 18:18:48 gram Exp $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Didier Jorand
 *
 * Some stuff from:
 * 
 * GXSNMP -- An snmp mangament application
 * Copyright (C) 1998 Gregory McLean & Jochen Friedrich
 * Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#define MAX_STRING_LEN 1024	/* TBC */

#include <glib.h>
#include <smi.h>

#include "packet.h"
#include "etypes.h"
#include "asn1.h"

#include "packet-snmp.h"

static int proto_snmp = -1;
static int proto_smux = -1;

static gint ett_snmp = -1;
static gint ett_smux = -1;
static gint ett_global = -1;
static gint ett_flags = -1;

static int hf_snmpv3_flags = -1;
static int hf_snmpv3_flags_auth = -1;
static int hf_snmpv3_flags_crypt = -1;
static int hf_snmpv3_flags_report = -1;

#define TH_AUTH   0x01
#define TH_CRYPT  0x02
#define TH_REPORT 0x04

static const true_false_string flags_set_truth = {
  "Set",
  "Not set"
};

#define UDP_PORT_SNMP		161
#define UDP_PORT_SNMP_TRAP	162
#define TCP_PORT_SMUX		199

/* Protocol version numbers */
#define SNMP_VERSION_1	0
#define SNMP_VERSION_2c	1
#define SNMP_VERSION_2u	2
#define SNMP_VERSION_3	3

static const value_string versions[] = {
	{ SNMP_VERSION_1,	"1" },
	{ SNMP_VERSION_2c,	"2C" },
	{ SNMP_VERSION_2u,	"2U" },
	{ SNMP_VERSION_3,	"3" },
	{ 0,			NULL },
};

/* PDU types */
#define SNMP_MSG_GET		0
#define SNMP_MSG_GETNEXT	1
#define SNMP_MSG_RESPONSE	2
#define SNMP_MSG_SET		3
#define SNMP_MSG_TRAP		4

#define SNMP_MSG_GETBULK	5
#define SNMP_MSG_INFORM		6
#define SNMP_MSG_TRAP2		7
#define SNMP_MSG_REPORT		8

static const value_string pdu_types[] = {
	{ SNMP_MSG_GET,		"GET" },
	{ SNMP_MSG_GETNEXT,	"GET-NEXT" },
	{ SNMP_MSG_SET,		"SET" },
	{ SNMP_MSG_RESPONSE,	"RESPONSE" },
	{ SNMP_MSG_TRAP, 	"TRAP-V1" },
	{ SNMP_MSG_GETBULK, 	"GETBULK" },
	{ SNMP_MSG_INFORM, 	"INFORM" },
	{ SNMP_MSG_TRAP2, 	"TRAP-V2" },
	{ SNMP_MSG_REPORT,	"REPORT" },
	{ 0,			NULL }
};

/* SMUX PDU types */
#define SMUX_MSG_OPEN 		0
#define SMUX_MSG_CLOSE		1
#define SMUX_MSG_RREQ		2
#define SMUX_MSG_RRSP		3
#define SMUX_MSG_SOUT		4

static const value_string smux_types[] = {
	{ SMUX_MSG_OPEN,	"Open" },
	{ SMUX_MSG_CLOSE,	"Close" },
	{ SMUX_MSG_RREQ,	"Registration Request" },
	{ SMUX_MSG_RRSP,	"Registration Response" },
	{ SMUX_MSG_SOUT,	"Commit Or Rollback" },
	{ 0,			NULL }
};

/* SMUX Closing causes */
#define SMUX_CLOSE_DOWN			0
#define SMUX_CLOSE_VERSION		1
#define SMUX_CLOSE_PACKET		2
#define SMUX_CLOSE_PROTOCOL		3
#define SMUX_CLOSE_INTERNAL		4
#define SMUX_CLOSE_NOAUTH		5

static const value_string smux_close[] = {
	{ SMUX_CLOSE_DOWN,	"Going down" },
	{ SMUX_CLOSE_VERSION,	"Unsupported Version" },
	{ SMUX_CLOSE_PACKET,	"Packet Format Error" },
	{ SMUX_CLOSE_PROTOCOL,	"Protocol Error" },
	{ SMUX_CLOSE_INTERNAL,	"Internal Error" },
	{ SMUX_CLOSE_NOAUTH,	"Unauthorized" },
	{ 0,			NULL }
};

/* SMUX Request codes */
#define SMUX_RREQ_DELETE		0
#define SMUX_RREQ_READONLY		1
#define SMUX_RREQ_READWRITE		2

static const value_string smux_rreq[] = {
	{ SMUX_RREQ_DELETE,	"Delete" },
	{ SMUX_RREQ_READONLY,	"Read Only" },
	{ SMUX_RREQ_READWRITE,	"Read Write" },
	{ 0,			NULL }
};

static const value_string smux_prio[] = {
	{ -1,				"Failure" },
	{ 0,				NULL }
};

/* SMUX SOut codes */
#define SMUX_SOUT_COMMIT		0
#define SMUX_SOUT_ROLLBACK		1

static const value_string smux_sout[] = {
	{ SMUX_SOUT_COMMIT,		"Commit" },
	{ SMUX_SOUT_ROLLBACK,		"Rollback" }
};

/* Error status values */
#define SNMP_ERR_NOERROR		0
#define SNMP_ERR_TOOBIG			1
#define SNMP_ERR_NOSUCHNAME		2
#define SNMP_ERR_BADVALUE		3
#define SNMP_ERR_READONLY		4
#define SNMP_ERR_GENERROR		5

#define SNMP_ERR_NOACCESS		6
#define SNMP_ERR_WRONGTYPE		7
#define SNMP_ERR_WRONGLENGTH		8
#define SNMP_ERR_WRONGENCODING		9
#define SNMP_ERR_WRONGVALUE		10
#define SNMP_ERR_NOCREATION		11
#define SNMP_ERR_INCONSISTENTVALUE	12
#define SNMP_ERR_RESOURCEUNAVAILABLE	13
#define SNMP_ERR_COMMITFAILED		14
#define SNMP_ERR_UNDOFAILED		15
#define SNMP_ERR_AUTHORIZATIONERROR	16
#define SNMP_ERR_NOTWRITABLE		17
#define SNMP_ERR_INCONSISTENTNAME	18

static const value_string error_statuses[] = {
	{ SNMP_ERR_NOERROR,		"NO ERROR" },
	{ SNMP_ERR_TOOBIG,		"TOOBIG" },
	{ SNMP_ERR_NOSUCHNAME,		"NO SUCH NAME" },
	{ SNMP_ERR_BADVALUE,		"BAD VALUE" },
	{ SNMP_ERR_READONLY,		"READ ONLY" },
	{ SNMP_ERR_GENERROR,		"GENERIC ERROR" },
	{ SNMP_ERR_NOACCESS,		"NO ACCESS" },
	{ SNMP_ERR_WRONGTYPE,		"WRONG TYPE" },
	{ SNMP_ERR_WRONGLENGTH,		"WRONG LENGTH" },
	{ SNMP_ERR_WRONGENCODING,	"WRONG ENCODING" },
	{ SNMP_ERR_WRONGVALUE,		"WRONG VALUE" },
	{ SNMP_ERR_NOCREATION,		"NO CREATION" },
	{ SNMP_ERR_INCONSISTENTVALUE,	"INCONSISTENT VALUE" },
	{ SNMP_ERR_RESOURCEUNAVAILABLE,	"RESOURCE UNAVAILABLE" },
	{ SNMP_ERR_COMMITFAILED,	"COMMIT FAILED" },
	{ SNMP_ERR_UNDOFAILED,		"UNDO FAILED" },
	{ SNMP_ERR_AUTHORIZATIONERROR,	"AUTHORIZATION ERROR" },
	{ SNMP_ERR_NOTWRITABLE,		"NOT WRITABLE" },
	{ SNMP_ERR_INCONSISTENTNAME,	"INCONSISTENT NAME" },
	{ 0,				NULL }
};

/* General SNMP V1 Traps */

#define SNMP_TRAP_COLDSTART		0
#define SNMP_TRAP_WARMSTART		1
#define SNMP_TRAP_LINKDOWN		2
#define SNMP_TRAP_LINKUP		3
#define SNMP_TRAP_AUTHFAIL		4
#define SNMP_TRAP_EGPNEIGHBORLOSS	5
#define SNMP_TRAP_ENTERPRISESPECIFIC	6

static const value_string trap_types[] = {
	{ SNMP_TRAP_COLDSTART,		"COLD START" },
	{ SNMP_TRAP_WARMSTART,		"WARM START" },
	{ SNMP_TRAP_LINKDOWN,		"LINK DOWN" },
	{ SNMP_TRAP_LINKUP,		"LINK UP" },
	{ SNMP_TRAP_AUTHFAIL,		"AUTHENTICATION FAILED" },
	{ SNMP_TRAP_EGPNEIGHBORLOSS,	"EGP NEIGHBORLOSS" },
	{ SNMP_TRAP_ENTERPRISESPECIFIC,	"ENTERPRISE SPECIFIC" },
	{ 0,				NULL }
};

/* SNMP Tags */

#define SNMP_IPA    0		/* IP Address */
#define SNMP_CNT    1		/* Counter (Counter32) */
#define SNMP_GGE    2		/* Gauge (Gauge32) */
#define SNMP_TIT    3		/* TimeTicks */
#define SNMP_OPQ    4		/* Opaque */
#define SNMP_NSP    5		/* NsapAddress */
#define SNMP_C64    6		/* Counter64 */
#define SNMP_U32    7		/* Uinteger32 */

#define SERR_NSO    0
#define SERR_NSI    1
#define SERR_EOM    2

/* SNMPv1 Types */

#define SNMP_NULL                0
#define SNMP_INTEGER             1    /* l  */
#define SNMP_OCTETSTR            2    /* c  */
#define SNMP_DISPLAYSTR          2    /* c  */
#define SNMP_OBJECTID            3    /* ul */
#define SNMP_IPADDR              4    /* uc */
#define SNMP_COUNTER             5    /* ul */
#define SNMP_GAUGE               6    /* ul */
#define SNMP_TIMETICKS           7    /* ul */
#define SNMP_OPAQUE              8    /* c  */

/* additional SNMPv2 Types */

#define SNMP_UINTEGER            5    /* ul */
#define SNMP_BITSTR              9    /* uc */
#define SNMP_NSAP               10    /* uc */
#define SNMP_COUNTER64          11    /* ul */
#define SNMP_NOSUCHOBJECT       12
#define SNMP_NOSUCHINSTANCE     13
#define SNMP_ENDOFMIBVIEW       14

typedef struct _SNMP_CNV SNMP_CNV;

struct _SNMP_CNV
{
  guint class;
  guint tag;
  gint  syntax;
  gchar *name;
};

static SNMP_CNV SnmpCnv [] =
{
  {ASN1_UNI, ASN1_NUL, SNMP_NULL,      "NULL"},
  {ASN1_UNI, ASN1_INT, SNMP_INTEGER,   "INTEGER"},
  {ASN1_UNI, ASN1_OTS, SNMP_OCTETSTR,  "OCTET STRING"},
  {ASN1_UNI, ASN1_OJI, SNMP_OBJECTID,  "OBJECTID"},
  {ASN1_APL, SNMP_IPA, SNMP_IPADDR,    "IPADDR"},
  {ASN1_APL, SNMP_CNT, SNMP_COUNTER,   "COUNTER"},  /* Counter32 */
  {ASN1_APL, SNMP_GGE, SNMP_GAUGE,     "GAUGE"},    /* Gauge32 == Unsigned32  */
  {ASN1_APL, SNMP_TIT, SNMP_TIMETICKS, "TIMETICKS"},
  {ASN1_APL, SNMP_OPQ, SNMP_OPAQUE,    "OPAQUE"},

/* SNMPv2 data types and errors */

  {ASN1_UNI, ASN1_BTS, SNMP_BITSTR,         "BITSTR"},
  {ASN1_APL, SNMP_C64, SNMP_COUNTER64,      "COUNTER64"},
  {ASN1_CTX, SERR_NSO, SNMP_NOSUCHOBJECT,   "NOSUCHOBJECT"},
  {ASN1_CTX, SERR_NSI, SNMP_NOSUCHINSTANCE, "NOSUCHINSTANCE"},
  {ASN1_CTX, SERR_EOM, SNMP_ENDOFMIBVIEW,   "ENDOFMIBVIEW"},
  {0,       0,         -1,                  NULL}
};

static gint
int_div(guchar *buffer, gint radix, gint *len, gint *nul)
{
  gint dividend, i, mod;

  *nul = 1;
  while (*len && (*buffer == 0))
    {
      for (i=1; i<*len; i++)
        buffer[i-1] = buffer[i];
      (*len)--;
    }
  if (!*len) 
    return 0;

  dividend = 0;
  mod = 0;
  for (i=0; i<*len; i++)
    {
      dividend = mod * 256 + buffer[i];
      mod = dividend % radix;
      buffer[i] = dividend / radix;
    }
  while (*len && (*buffer == 0))
    {
      for (i=1; i<*len; i++)
        buffer[i-1] = buffer[i];
      (*len)--;
    }
  if (*len) *nul=0;
  return mod;
}

/**
 * print_radix:
 * @buffer: integer value
 * @len: length of integer
 * @radix: base
 * @shift: decimal shift
 *
 */
static gchar *
print_radix(gpointer buffer, gint len, gint radix, gboolean sign, gint shift)
{
  guchar *result, *temp;
  gint size, neg, i, nul;
  gchar digit[] = "0123456789ABCDEF";

  if (radix<2) radix=2;
  if (radix>16) radix=16;

  size = len*8 + 2; /* worst case + sign + trailing 0 */
  if (shift) size++; /* decimal point */
  if (size<shift+4) size=shift+4; /* additional leading 0 */

  result = g_malloc(size+1);
  temp = g_malloc(len);

  neg = 0;

  g_memmove(temp, buffer, len);

  if (sign && (temp[0] & 0x80))
    {
      neg = 1;
      for (i=0; i<len; i++)
	temp[i] = ~temp[i];
      i=len-1;
      while (i>=0)
	{
	  temp[i]++;
	  if (temp[i]) break;
	  i--;
	}
    }
  i=size;
  result[i--] = '\0';
  nul=0;
  while ((!nul || (shift && size-i-3 < shift)) && (i>=0))
    {
      if ((shift == size-i-1) && shift)
	result[i--] = '.';
      else
	result[i--] = digit[int_div(temp, radix, &len, &nul)];
    }
  if (i && neg)
    result[i--] = '-';
  g_free(temp);
  temp = g_strdup(result+i+1);
  g_free(result);
  return temp;
}
/** 
 * format_integer:
 * @buffer: integer value as received from SNMP node
 * @len: length of integer
 * @format: Format as taken from DISPLAY HINT parameter of MIB file.
 *
 * Formats an INTEGER value according to a provided DISPLAY HINT.
 *
 * When the syntax has an underlying primitive type of INTEGER, the hint
 * consists of an integer-format specification, containing two parts.
 * The first part is a single character suggesting a display format,
 * either: `x' for hexadecimal, or `d' for decimal, or `o' for octal, or
 * `b' for binary.  For all types, when rendering the value, leading
 * zeros are omitted, and for negative values, a minus sign is rendered
 * immediately before the digits.  The second part is always omitted for
 * `x', `o' and `b', and need not be present for `d'.  If present, the
 * second part starts with a hyphen and is followed by a decimal number,
 * which defines the implied decimal point when rendering the value.
 * 
 * Return value: Formatted string.
 **/
static gchar*
format_integer(gpointer buffer, gint len, gboolean sign, gchar *format)
{
  gint shift;

  shift = 0;
  switch(format[0])
    {
      case 'x':
	return print_radix(buffer, len, 16, sign, 0);
	break;
      case 'o':
	return print_radix(buffer, len, 8, sign, 0);
	break;
      case 'b':
	return print_radix(buffer, len, 2, sign, 0);
	break;
      case 'd':
	if (format[1] == '-')
	  shift = atoi(&(format[2]));
	if (shift < 0)
	  shift = 0;
      default:
	return print_radix(buffer, len, 10, sign, shift);
    }
}

/** 
 * format_string:
 * @buffer: octet string as received from SNMP node
 * @len: length of octet string
 * @format: Format as taken from DISPLAY HINT parameter of MIB file.
 *
 * Formats an OCTET STRING value according to a provided DISPLAY HINT.
 *
 * When the syntax has an underlying primitive type of OCTET STRING, the
 * hint consists of one or more octet-format specifications.  Each
 * specification consists of five parts, with each part using and
 * removing zero or more of the next octets from the value and producing
 * the next zero or more characters to be displayed.  The octets within
 * the value are processed in order of significance, most significant
 * first.
 *
 * The five parts of a octet-format specification are:
 *
 * (1)  the (optional) repeat indicator; if present, this part is a `*',
 *      and indicates that the current octet of the value is to be used as
 *      the repeat count.  The repeat count is an unsigned integer (which
 *      may be zero) which specifies how many times the remainder of this
 *      octet-format specification should be successively applied.  If the
 *      repeat indicator is not present, the repeat count is one.
 *
 * (2)  the octet length: one or more decimal digits specifying the number
 *      of octets of the value to be used and formatted by this octet-
 *      specification.  Note that the octet length can be zero.  If less
 *      than this number of octets remain in the value, then the lesser
 *      number of octets are used.
 *
 * (3)  the display format, either:  `x' for hexadecimal, `d' for decimal,
 *      `o' for octal, `a' for ascii, or `t' for UTF-8.  If the octet
 *      length part is greater than one, and the display format part refers
 *      to a numeric format, then network-byte ordering (big-endian
 *      encoding) is used interpreting the octets in the value.  The octets
 *      processed by the `t' display format do not necessarily form an
 *      integral number of UTF-8 characters.  Trailing octets which do not
 *      form a valid UTF-8 encoded character are discarded.
 *
 * (4)  the (optional) display separator character; if present, this part
 *      is a single character which is produced for display after each
 *      application of this octet-specification; however, this character is
 *      not produced for display if it would be immediately followed by the
 *      display of the repeat terminator character for this octet-
 *      specification.  This character can be any character other than a
 *      decimal digit and a `*'.
 *
 * (5)  the (optional) repeat terminator character, which can be present
 *      only if the display separator character is present and this octet-
 *      specification begins with a repeat indicator; if present, this part
 *      is a single character which is produced after all the zero or more
 *      repeated applications (as given by the repeat count) of this
 *      octet-specification.  This character can be any character other
 *      than a decimal digit and a `*'.
 *
 * Output of a display separator character or a repeat terminator
 * character is suppressed if it would occur as the last character of
 * the display.
 *
 * If the octets of the value are exhausted before all the octet-format
 * specification have been used, then the excess specifications are
 * ignored.  If additional octets remain in the value after interpreting
 * all the octet-format specifications, then the last octet-format
 * specification is re-interpreted to process the additional octets,
 * until no octets remain in the value.
 * 
 * Return value: Formatted string.
 **/
static gchar*
format_string(gpointer buffer, gint len, gchar *format)
{
  int repeat, count, size;
  gchar seperator, terminator, formtype;
  gchar *ptr;
  guchar *bufptr;
  gchar *result;
  int pos;
  gchar *numstr;
  int numstrlen;

  if (!len)
    return g_strdup("NULL");

  ptr = format;
  bufptr = (guchar *)buffer;

  size = len * 3 * strlen(format);
  result = g_malloc(size);

  pos = 0;

  while (len)
    {
    /* Read format string */
      if (!*ptr)
        ptr = format;
      repeat = 1;
      count = 0;
      seperator = '\0';
      terminator = '\0';
      formtype = '\0';
      if (*ptr == '*')
        {
          repeat = *bufptr++;
	  len--;
	  ptr++;
	}
      while (*ptr >= '0' && *ptr <= '9')
        {
	  count = 10 * count + (*ptr++ & 0xf);
	}
      if (count == 0)
        count = 1;

      if (*ptr && (*ptr != '*') && ((*ptr < '0') || (*ptr > '9')))
        formtype = *ptr++;
      if (*ptr && (*ptr != '*') && ((*ptr < '0') || (*ptr > '9')))
        seperator = *ptr++;
      if (*ptr && (*ptr != '*') && ((*ptr < '0') || (*ptr > '9')))
        terminator = *ptr++;

      while(repeat && len)
        {
          if (count > len) count = len;
          switch (formtype)
            {
	      case 'x':
	        numstr = print_radix(bufptr, count, 16, FALSE, 0);
	        numstrlen = strlen(numstr);
		g_assert(pos + numstrlen < size);
	        strcpy(result + pos, numstr);
	        pos += numstrlen;
	        g_free(numstr);
	        break;
	      case 'd':
	        numstr = print_radix(bufptr, count, 10, FALSE, 0);
	        numstrlen = strlen(numstr);
		g_assert(pos + numstrlen < size);
	        strcpy(result + pos, numstr);
	        pos += numstrlen;
	        g_free(numstr);
	        break;
	      case 'o':
	        numstr = print_radix(bufptr, count, 8, FALSE, 0);
	        numstrlen = strlen(numstr);
		g_assert(pos + numstrlen < size);
	        strcpy(result + pos, numstr);
	        pos += numstrlen;
	        g_free(numstr);
	        break;
	      case 'a':
	        g_assert(pos + count < size);
	        g_memmove(result + pos, bufptr, count);
	        pos += count;
		break;
            }
	   
          len = len - count;
          bufptr = bufptr + count;
	  repeat--;
	  if (!repeat && len)
	    {
	      if (terminator)
	        {
		  g_assert(pos + 1 < size);
	          result[pos++] = terminator;
	        }
	      else if (seperator)
	        {
		  g_assert(pos + 1 < size);
	          result[pos++] = seperator;
	        }
            }
	  else if (seperator && len)
	    {
	      g_assert(pos + 1 < size);
	      result[pos++] = seperator;
	    }
	}
    }
  result[pos] = 0;
  ptr = g_strdup(result);
  g_free(result);
  return ptr;
}

/*
 * NAME:        g_snmp_tag_cls2syntax
 * SYNOPSIS:    gboolean g_snmp_tag_cls2syntax
 *                  (
 *                      guint    tag,
 *                      guint    cls,
 *                      gushort *syntax
 *                  )
 * DESCRIPTION: Converts ASN1 tag and class to Syntax tag and name.
 *              See SnmpCnv for conversion.
 * RETURNS:     name on success, NULL on failure
 */

static gchar *
snmp_tag_cls2syntax ( guint tag, guint cls, gushort *syntax)
{
    SNMP_CNV *cnv;

    cnv = SnmpCnv;
    while (cnv->syntax != -1)
    {
        if (cnv->tag == tag && cnv->class == cls)
        {
            *syntax = cnv->syntax;
            return cnv->name;
        }
        cnv++;
    }
    return NULL;
}

static void
dissect_snmp_parse_error(const u_char *pd, int offset, frame_data *fd,
		   proto_tree *tree, const char *field_name, int ret)
{
	const gchar *errstr;

	if (check_col(fd, COL_INFO)) {
		switch (ret) {

		case ASN1_ERR_EMPTY:
			errstr = "Ran out of data";
			break;

		case ASN1_ERR_EOC_MISMATCH:
			errstr = "EOC mismatch";
			break;

		case ASN1_ERR_WRONG_TYPE:
			errstr = "Wrong type for that item";
			break;

		case ASN1_ERR_LENGTH_NOT_DEFINITE:
			errstr = "Length was indefinite";
			break;

		case ASN1_ERR_LENGTH_MISMATCH:
			errstr = "Length mismatch";
			break;

		case ASN1_ERR_WRONG_LENGTH_FOR_TYPE:
			errstr = "Wrong length for that item's type";
			break;

		default:
			errstr = "Unknown error";
			break;
		}
		col_add_fstr(fd, COL_INFO,
		    "ERROR: Couldn't parse %s: %s", field_name, errstr);
	}

	dissect_data(pd, offset, fd, tree);
}

static void
dissect_snmp_error(const u_char *pd, int offset, frame_data *fd,
		   proto_tree *tree, const char *message)
{
	if (check_col(fd, COL_INFO))
		col_add_str(fd, COL_INFO, message);

	dissect_data(pd, offset, fd, tree);
}

static void
format_oid(gchar *buf, subid_t *oid, guint oid_length)
{
	int i;
	int len;

	len = sprintf(buf, "%lu", (unsigned long)oid[0]);
	buf += len;
	for (i = 1; i < oid_length;i++) {
		len = sprintf(buf, ".%lu", (unsigned long)oid[i]);
		buf += len;
	}
}

static gchar*
format_var(gushort vb_type, guint8 *vb_octet_string, guint vb_length,
    SmiNode *node)
{
	SmiType *typ;
	gchar *string;
	char *format;
	int i;
	SmiInteger32 integer32;
	SmiUnsigned32 unsigned32;
	SmiNamedNumber *named_number;
	gchar *enumstring;

	typ = smiGetNodeType(node);
	if (typ != NULL)
		format = typ->format;
	else
		format = NULL;

	if (format == NULL) {
		switch (vb_type) {

		case SNMP_INTEGER:
		case SNMP_COUNTER:
		case SNMP_GAUGE:
		case SNMP_TIMETICKS:
		case SNMP_COUNTER64:
			format = "d";
			break;

		case SNMP_OCTETSTR:
		case SNMP_OPAQUE:
		case SNMP_NSAP:
		case SNMP_BITSTR:
			format = "1a";
			for (i = 0; i < vb_length; i++) {
				if (!(isprint(vb_octet_string[i])
				    || isspace(vb_octet_string[i]))) {
					/*
					 * Oh, dear, it's binary.
					 */
					format = "1x ";
					break;
				}
			}
			break;

		case SNMP_IPADDR:
			format = "d.";
			break;

		default:
			format = NULL;
			break;
		}
	}

	switch (vb_type) {

	case SNMP_INTEGER:
		string = format_integer(vb_octet_string, vb_length, TRUE,
		    format);
		if (typ != NULL && vb_length <= 4) {
			/*
			 * OK, we have a module, and the number fits in
			 * an "SmiInteger32".  Check if there's an enum
			 * for this.
			 *
			 * XXX - handle "SmiInteger64"?
			 */
			integer32 = 0;
			i = 0;
			while (vb_length != 0) {
				integer32 =
				    (integer32 << 8) | vb_octet_string[i];
				i++;
				vb_length--;
			}
			for (named_number = smiGetFirstNamedNumber(typ);
			    named_number != NULL;
			    named_number = smiGetNextNamedNumber(named_number)) {
				if (named_number->value.basetype == SMI_BASETYPE_INTEGER32
				    && named_number->value.value.integer32 == integer32)
					break;
			}
			if (named_number != NULL) {
				enumstring = g_malloc(
				    strlen(named_number->name) + 1 +
				    strlen(string) + 1 + 1);
				sprintf(enumstring, "%s(%s)",
				    named_number->name, string);
				g_free(string);
				string = enumstring;
			}
		}
		break;

	case SNMP_COUNTER:
	case SNMP_GAUGE:
	case SNMP_TIMETICKS:
	case SNMP_COUNTER64:
		string = format_integer(vb_octet_string, vb_length, FALSE,
		    format);
		if (typ != NULL && vb_length <= 4) {
			/*
			 * OK, we have a module, and the number fits in
			 * an "SmiUnsigned32".  Check if there's an enum
			 * for this.
			 *
			 * XXX - handle "SmiUnsigned64"?
			 */
			unsigned32 = 0;
			i = 0;
			while (vb_length != 0) {
				unsigned32 =
				    (unsigned32 << 8) | vb_octet_string[i];
				i++;
				vb_length--;
			}
			for (named_number = smiGetFirstNamedNumber(typ);
			    named_number != NULL;
			    named_number = smiGetNextNamedNumber(named_number)) {
				if (named_number->value.basetype == SMI_BASETYPE_UNSIGNED32
				    && named_number->value.value.unsigned32 == unsigned32)
					break;
			}
			if (named_number != NULL) {
				enumstring = g_malloc(
				    strlen(named_number->name) + 1 +
				    strlen(string) + 1 + 1);
				sprintf(enumstring, "%s(%s)",
				    named_number->name, string);
				g_free(string);
				string = enumstring;
			}
		}
		break;

	case SNMP_OCTETSTR:
	case SNMP_IPADDR:
	case SNMP_OPAQUE:
	case SNMP_NSAP:
	case SNMP_BITSTR:
		string = format_string(vb_octet_string, vb_length, format);
		break;

	default:
		string = NULL;
		break;
	}
	return string;
}

static int
snmp_variable_decode(proto_tree *snmp_tree, subid_t *variable_oid,
    guint variable_oid_length, ASN1_SCK *asn1, int offset, guint *lengthp, 
    SmiNode *node)
{
	const guchar *start;
	guint length;
	gboolean def;
	guint vb_length;
	gushort vb_type;
	gchar *vb_type_name;
	int ret;
	guint cls, con, tag;

	guint8 *vb_octet_string;

	subid_t *vb_oid;
	guint vb_oid_length;

	gchar *vb_display_string;

	SmiNode *oid_node;
	SmiModule *oid_module;
	gchar oid_string[MAX_STRING_LEN]; /* TBC */
	gchar vb_oid_string[MAX_STRING_LEN]; /* TBC */

	/* parse the type of the object */
	start = asn1->pointer;
	ret = asn1_header_decode (asn1, &cls, &con, &tag, &def, &vb_length);
	if (ret != ASN1_ERR_NOERROR)
		return ret;
	if (!def)
		return ASN1_ERR_LENGTH_NOT_DEFINITE;

	/* Convert the class, constructed flag, and tag to a type. */
	vb_type_name = snmp_tag_cls2syntax(tag, cls, &vb_type);
	if (vb_type_name == NULL) {
		/*
		 * Unsupported type.
		 * Dissect the value as an opaque string of octets.
		 */
		vb_type_name = "unsupported type";
		vb_type = SNMP_OPAQUE;
	}

	/* parse the value */
	switch (vb_type) {

	case SNMP_INTEGER:
	case SNMP_COUNTER:
	case SNMP_GAUGE:
	case SNMP_TIMETICKS:
	case SNMP_COUNTER64:
	case SNMP_OCTETSTR:
	case SNMP_IPADDR:
	case SNMP_OPAQUE:
	case SNMP_NSAP:
	case SNMP_BITSTR:
		ret = asn1_octet_string_value_decode (asn1, vb_length,
		    &vb_octet_string);
		if (ret != ASN1_ERR_NOERROR)
			return ret;
		length = asn1->pointer - start;
		if (snmp_tree) {
			vb_display_string = format_var(vb_type, vb_octet_string,
			    vb_length, node);
			proto_tree_add_text(snmp_tree, offset, length,
			    "Value: %s: %s", vb_type_name,
			    vb_display_string);
			g_free(vb_display_string);
		}
		g_free(vb_octet_string);
		break;

	case SNMP_NULL:
		ret = asn1_null_decode (asn1, vb_length);
		if (ret != ASN1_ERR_NOERROR)
			return ret;
		length = asn1->pointer - start;
		if (snmp_tree) {
			proto_tree_add_text(snmp_tree, offset, length,
			    "Value: %s", vb_type_name);
		}
		break;

	case SNMP_OBJECTID:
		ret = asn1_oid_value_decode (asn1, vb_length, &vb_oid,
		    &vb_oid_length);
		if (ret != ASN1_ERR_NOERROR)
			return ret;
		length = asn1->pointer - start;
		if (snmp_tree) {
			format_oid(oid_string, vb_oid, vb_oid_length);
			oid_node = smiGetNodeByOID(vb_oid_length,
				       (SmiSubid *)vb_oid);
			if (oid_node) {
				oid_module = smiGetNodeModule(oid_node);
				sprintf(vb_oid_string, " (%s:%s)",
				    oid_module->name, oid_node->name);
			} else
				strcpy(vb_oid_string, "");
			proto_tree_add_text(snmp_tree, offset, length,
			    "Value: %s: %s%s", vb_type_name, oid_string,
			    vb_oid_string);
		}
		g_free(vb_oid);
		break;

	case SNMP_NOSUCHOBJECT:
		length = asn1->pointer - start;
		if (snmp_tree) {
			proto_tree_add_text(snmp_tree, offset, length,
			    "Value: %s: no such object", vb_type_name);
		}
		break;

	case SNMP_NOSUCHINSTANCE:
		length = asn1->pointer - start;
		if (snmp_tree) {
			proto_tree_add_text(snmp_tree, offset, length,
			    "Value: %s: no such instance", vb_type_name);
		}
		break;

	case SNMP_ENDOFMIBVIEW:
		length = asn1->pointer - start;
		if (snmp_tree) {
			proto_tree_add_text(snmp_tree, offset, length,
			    "Value: %s: end of mib view", vb_type_name);
		}
		break;

	default:
		g_assert_not_reached();
		return ASN1_ERR_WRONG_TYPE;
	}
	*lengthp = length;
	return ASN1_ERR_NOERROR;
}

void
dissect_common_pdu(const u_char *pd, int offset, frame_data *fd,
    proto_tree *tree, ASN1_SCK asn1, guint pdu_type, const guchar *start)
{
	gboolean def;
	guint length;
	guint sequence_length;

	guint32 request_id;

	guint32 error_status;

	guint32 error_index;

	char *pdu_type_string;

	subid_t *enterprise;
	guint enterprise_length;

	guint8 *agent_address;
	guint agent_address_length;

	guint32 trap_type;

	guint32 specific_type;

	guint timestamp;
	guint timestamp_length;

	gchar oid_string[MAX_STRING_LEN]; /* TBC */

	guint variable_bindings_length;

	int vb_index;
	guint variable_length;
	subid_t *variable_oid;
	guint variable_oid_length;
	gchar vb_oid_string[MAX_STRING_LEN]; /* TBC */

	int ret;
	guint cls, con, tag;

	SmiNode *node;
	SmiModule *module;

	pdu_type_string = val_to_str(pdu_type, pdu_types,
	    "Unknown PDU type %#x");
	if (check_col(fd, COL_INFO))
		col_add_str(fd, COL_INFO, pdu_type_string);
	length = asn1.pointer - start;
	if (tree) {
		proto_tree_add_text(tree, offset, length,
		    "PDU type: %s", pdu_type_string);
	}
	offset += length;

	/* get the fields in the PDU preceeding the variable-bindings sequence */
	switch (pdu_type) {

	case SNMP_MSG_GET:
	case SNMP_MSG_GETNEXT:
	case SNMP_MSG_RESPONSE:
	case SNMP_MSG_SET:
	case SNMP_MSG_GETBULK:
	case SNMP_MSG_INFORM:
	case SNMP_MSG_TRAP2:
	case SNMP_MSG_REPORT:
		/* request id */
		ret = asn1_uint32_decode (&asn1, &request_id, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "request ID", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(tree, offset, length,
			    "Request Id: %#x", request_id);
		}
		offset += length;
		
		/* error status, or getbulk non-repeaters */
		ret = asn1_uint32_decode (&asn1, &error_status, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    (pdu_type == SNMP_MSG_GETBULK) ? "non-repeaters"
			    				   : "error status",
			    ret);
			return;
		}
		if (tree) {
			if (pdu_type == SNMP_MSG_GETBULK) {
				proto_tree_add_text(tree, offset, length,
				    "Non-repeaters: %u", error_status);
			} else {
				proto_tree_add_text(tree, offset, length,
				    "Error Status: %s",
				    val_to_str(error_status, error_statuses,
				      "Unknown (%d)"));
			}
		}
		offset += length;

		/* error index, or getbulk max-repetitions */
		ret = asn1_uint32_decode (&asn1, &error_index, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    (pdu_type == SNMP_MSG_GETBULK) ? "max repetitions"
			    				   : "error index",
			    ret);
			return;
		}
		if (tree) {
			if (pdu_type == SNMP_MSG_GETBULK) {
				proto_tree_add_text(tree, offset, length,
				    "Max repetitions: %u", error_index);
			} else {
				proto_tree_add_text(tree, offset, length,
				    "Error Index: %u", error_index);
			}
		}
		offset += length;
		break;

	case SNMP_MSG_TRAP:
		/* enterprise */
		ret = asn1_oid_decode (&asn1, &enterprise, &enterprise_length,
		    &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "enterprise OID", ret);
			return;
		}
		if (tree) {
			format_oid(oid_string, enterprise, enterprise_length);
			proto_tree_add_text(tree, offset, length,
			    "Enterprise: %s", oid_string);
		}
		g_free(enterprise);
		offset += length;

		/* agent address */
		start = asn1.pointer;
		ret = asn1_header_decode (&asn1, &cls, &con, &tag,
		    &def, &agent_address_length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "agent address", ret);
			return;
		}
		if (!((cls == ASN1_APL && con == ASN1_PRI && tag == SNMP_IPA) ||
		    (cls == ASN1_UNI && con == ASN1_PRI && tag == ASN1_OTS))) {
			/* GXSNMP 0.0.15 says the latter is "needed for
			   Banyan" */
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "agent_address", ASN1_ERR_WRONG_TYPE);
			return;
		}
		if (agent_address_length != 4) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "agent_address", ASN1_ERR_WRONG_LENGTH_FOR_TYPE);
			return;
		}
		ret = asn1_octet_string_value_decode (&asn1,
		    agent_address_length, &agent_address);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "agent address", ret);
			return;
		}
		length = asn1.pointer - start;
		if (tree) {
			proto_tree_add_text(tree, offset, agent_address_length,
			    "Agent address: %s", ip_to_str(agent_address));
		}
		g_free(agent_address);
		offset += length;
		
	        /* generic trap type */
		ret = asn1_uint32_decode (&asn1, &trap_type, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "generic trap type", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(tree, offset, length,
			    "Trap type: %s",
			    val_to_str(trap_type, trap_types, "Unknown (%u)"));
		}		
		offset += length;
		
	        /* specific trap type */
		ret = asn1_uint32_decode (&asn1, &specific_type, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "specific trap type", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(tree, offset, length,
			    "Specific trap type: %u (%#x)",
			    specific_type, specific_type);
		}		
		offset += length;
		
	        /* timestamp */
		start = asn1.pointer;
		ret = asn1_header_decode (&asn1, &cls, &con, &tag,
		    &def, &timestamp_length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "timestamp", ret);
			return;
		}
		if (!((cls == ASN1_APL && con == ASN1_PRI && tag == SNMP_TIT) ||
		    (cls == ASN1_UNI && con == ASN1_PRI && tag == ASN1_INT))) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "timestamp", ASN1_ERR_WRONG_TYPE);
			return;
		}
		ret = asn1_uint32_value_decode(&asn1, timestamp_length,
		    &timestamp);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "timestamp", ret);
			return;
		}
		length = asn1.pointer - start;
		if (tree) {
			proto_tree_add_text(tree, offset, length,
			    "Timestamp: %u", timestamp);
		}		
		offset += length;
		break;
	}

	/* variable bindings */
	/* get header for variable-bindings sequence */
	ret = asn1_sequence_decode(&asn1, &variable_bindings_length, &length);
	if (ret != ASN1_ERR_NOERROR) {
		dissect_snmp_parse_error(pd, offset, fd, tree,
			"variable bindings header", ret);
		return;
	}
	offset += length;

	/* loop on variable bindings */
	vb_index = 0;
	while (variable_bindings_length > 0) {
		vb_index++;
		sequence_length = 0;

		/* parse type */
		ret = asn1_sequence_decode(&asn1, &variable_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
				"variable binding header", ret);
			return;
		}
		sequence_length += length;

		/* parse object identifier */
		ret = asn1_oid_decode (&asn1, &variable_oid,
		    &variable_oid_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "variable binding OID", ret);
			return;
		}
		sequence_length += length;

		node = smiGetNodeByOID(variable_oid_length, 
				       (SmiSubid *)variable_oid);
		if (tree) {
			format_oid(oid_string, variable_oid,
			    variable_oid_length);
			if (node) {
				module = smiGetNodeModule(node);
				sprintf(vb_oid_string, " (%s:%s)",
				    module->name, node->name);
			} else
				strcpy(vb_oid_string, "");
			proto_tree_add_text(tree, offset, sequence_length,
			    "Object identifier %d: %s%s", vb_index,
			    oid_string, vb_oid_string);
		}
		offset += sequence_length;
		variable_bindings_length -= sequence_length;
				
		/* Parse the variable's value */
		ret = snmp_variable_decode(tree, variable_oid,
		    variable_oid_length, &asn1, offset, &length, node);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "variable", ret);
			return;
		}
		offset += length;
		variable_bindings_length -= length;
	}
}

void
dissect_snmp_pdu(const u_char *pd, int offset, frame_data *fd,
    proto_tree *tree, char *proto_name, int proto, gint ett)
{
	ASN1_SCK asn1;
	const guchar *start;
	gboolean def;
	guint length;
	guint message_length;
	guint global_length;

	guint32 version;
	guint32 msgid;
	guint32 msgmax;
	guint32 msgsec;

	guchar *msgflags;
	guchar *community;
	guchar *secparm;
	guchar *cengineid;
	guchar *cname;
	int msgflags_length;
	int community_length;
	int secparm_length;
	int cengineid_length;
	int cname_length;

	guint pdu_type;
	guint pdu_length;

	proto_tree *snmp_tree = NULL;
	proto_tree *global_tree = NULL;
	proto_tree *flags_tree = NULL;
	proto_item *item = NULL;
	int ret;
	guint cls, con;

	if (check_col(fd, COL_PROTOCOL))
		col_add_str(fd, COL_PROTOCOL, proto_name);

	if (tree) {
		item = proto_tree_add_item(tree, proto, offset, END_OF_FRAME, NULL);
		snmp_tree = proto_item_add_subtree(item, ett);
	}

	/* NOTE: we have to parse the message piece by piece, since the
	 * capture length may be less than the message length: a 'global'
	 * parsing is likely to fail.
	 */
	/* parse the SNMP header */
	asn1_open(&asn1, &pd[offset], END_OF_FRAME);
	ret = asn1_sequence_decode(&asn1, &message_length, &length);
	if (ret != ASN1_ERR_NOERROR) {
		dissect_snmp_parse_error(pd, offset, fd, tree,
			"message header", ret);
		return;
	}
	offset += length;

	ret = asn1_uint32_decode (&asn1, &version, &length);
	if (ret != ASN1_ERR_NOERROR) {
		dissect_snmp_parse_error(pd, offset, fd, tree, "version number",
		    ret);
		return;
	}
	if (snmp_tree) {
		proto_tree_add_text(snmp_tree, offset, length,
		    "Version: %s",
		    val_to_str(version, versions, "Unknown version %#x"));
	}
	offset += length;


	switch (version) {
	case SNMP_VERSION_1:
	case SNMP_VERSION_2c:
		ret = asn1_octet_string_decode (&asn1, &community, 
		    &community_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree, 
			    "community", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(snmp_tree, offset, length,
			    "Community: %.*s", community_length, community);
		}
		g_free(community);
		offset += length;
		break;
	case SNMP_VERSION_2u:
		/* FIXME */
		break;
	case SNMP_VERSION_3:
		ret = asn1_sequence_decode(&asn1, &global_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
				"message global header", ret);
			return;
		}
		if (snmp_tree) {
			item = proto_tree_add_text(snmp_tree, offset,
			    global_length + length, "Message Global Header");
			global_tree = proto_item_add_subtree(item, ett_global);
			proto_tree_add_text(global_tree, offset, length,
			    "Message Global Header Length: %d", global_length);
		}
		offset += length;
		ret = asn1_uint32_decode (&asn1, &msgid, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree, 
			    "message id", ret);
			return;
		}
		if (global_tree) {
			proto_tree_add_text(global_tree, offset, length,
			    "Message ID: %d", msgid);
		}
		offset += length;
		ret = asn1_uint32_decode (&asn1, &msgmax, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree, 
			    "message max size", ret);
			return;
		}
		if (global_tree) {
			proto_tree_add_text(global_tree, offset, length,
			    "Message Max Size: %d", msgmax);
		}
		offset += length;
		ret = asn1_octet_string_decode (&asn1, &msgflags, 
		    &msgflags_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree, 
			    "message flags", ret);
			return;
		}
		if (msgflags_length != 1) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "message flags wrong length", ret);
			g_free(msgflags);
			return;
		}
		if (global_tree) {
			item = proto_tree_add_uint_format(global_tree,
			    hf_snmpv3_flags, offset, length, msgflags[0],
			    "Flags: 0x%02x", msgflags[0]);
			flags_tree = proto_item_add_subtree(item, ett_flags);
			proto_tree_add_item(flags_tree, hf_snmpv3_flags_report,
			    offset, length, msgflags[0]);
			proto_tree_add_item(flags_tree, hf_snmpv3_flags_crypt,
			    offset, length, msgflags[0]);
			proto_tree_add_item(flags_tree, hf_snmpv3_flags_auth,
			    offset, length, msgflags[0]);
		}
		g_free(msgflags);
		offset += length;
		ret = asn1_uint32_decode (&asn1, &msgsec, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree, 
			    "message security model", ret);
			return;
		}
		if (global_tree) {
			proto_tree_add_text(global_tree, offset, length,
			    "Message Security Model: %d", msgsec);
		}
		offset += length;
		ret = asn1_octet_string_decode (&asn1, &secparm, 
		    &secparm_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree, 
			    "Message Security Parameters", ret);
			return;
		}
		if (snmp_tree) {
			proto_tree_add_text(snmp_tree, offset, length,
			    "Message Security Parameters: %s", secparm);
		}
		g_free(secparm);
		offset += length;
		/* PDU starts here. Fix decoding of encrypted PDU */
		ret = asn1_sequence_decode(&asn1, &global_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
				"message global header", ret);
			return;
		}
		offset += length;
		ret = asn1_octet_string_decode (&asn1, &cengineid, 
		    &cengineid_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree, 
			    "context engine id", ret);
			return;
		}
		if (snmp_tree) {
			proto_tree_add_text(snmp_tree, offset, length,
			    "Context Engine ID: %s", cengineid);
		}
		g_free(cengineid);
		offset += length;
		ret = asn1_octet_string_decode (&asn1, &cname, 
		    &cname_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree, 
			    "context name", ret);
			return;
		}
		if (snmp_tree) {
			proto_tree_add_text(snmp_tree, offset, length,
			    "Context Name: %s", cname);
		}
		g_free(cname);
		offset += length;
		break;
	default:
		dissect_snmp_error(pd, offset, fd, tree,
		    "PDU for unknown version of SNMP");
		return;
	}

	start = asn1.pointer;
	ret = asn1_header_decode (&asn1, &cls, &con, &pdu_type, &def,
	    &pdu_length);
	if (ret != ASN1_ERR_NOERROR) {
		dissect_snmp_parse_error(pd, offset, fd, tree,
		    "PDU type", ret);
		return;
	}
	if (cls != ASN1_CTX || con != ASN1_CON) {
		dissect_snmp_parse_error(pd, offset, fd, tree,
		    "PDU type", ASN1_ERR_WRONG_TYPE);
		return;
	}
	dissect_common_pdu(pd, offset, fd, snmp_tree, asn1, pdu_type, start);
}

void
dissect_smux_pdu(const u_char *pd, int offset, frame_data *fd,
    proto_tree *tree, char *proto_name, int proto, gint ett)
{
	ASN1_SCK asn1;
	const guchar *start;
	gboolean def;
	guint length;

	guint pdu_type;
	char *pdu_type_string;
	guint pdu_length;

	guint32 version;
	guint32 cause;
	guint32 priority;
	guint32 operation;
	guint32 commit;

	guchar *password;
	int password_length;

	guchar *application;
	int application_length;

	subid_t *regid;
	guint regid_length;

	gchar oid_string[MAX_STRING_LEN]; /* TBC */

	proto_tree *smux_tree = NULL;
	proto_item *item = NULL;
	int ret;
	guint cls, con;

	if (check_col(fd, COL_PROTOCOL))
		col_add_str(fd, COL_PROTOCOL, proto_name);

	if (tree) {
		item = proto_tree_add_item(tree, proto, offset, END_OF_FRAME, NULL);
		smux_tree = proto_item_add_subtree(item, ett);
	}

	/* NOTE: we have to parse the message piece by piece, since the
	 * capture length may be less than the message length: a 'global'
	 * parsing is likely to fail.
	 */
	/* parse the SNMP header */
	asn1_open(&asn1, &pd[offset], END_OF_FRAME);
	start = asn1.pointer;
	ret = asn1_header_decode (&asn1, &cls, &con, &pdu_type, &def,
	    &pdu_length);
	if (ret != ASN1_ERR_NOERROR) {
		dissect_snmp_parse_error(pd, offset, fd, tree,
		    "PDU type", ret);
		return;
	}
	/* Dissect SMUX here */
	if (cls == ASN1_APL && con == ASN1_CON && pdu_type == SMUX_MSG_OPEN) {
		pdu_type_string = val_to_str(pdu_type, smux_types,
		    "Unknown PDU type %#x");
		if (check_col(fd, COL_INFO))
			col_add_str(fd, COL_INFO, pdu_type_string);
		length = asn1.pointer - start;
		if (tree) {
			proto_tree_add_text(smux_tree, offset, length,
			    "PDU type: %s", pdu_type_string);
		}
		offset += length;
		ret = asn1_uint32_decode (&asn1, &version, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "version", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(smux_tree, offset, length,
			    "Version: %d", version);
		}
		offset += length;

		ret = asn1_oid_decode (&asn1, &regid, &regid_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "registration OID", ret);
			return;
		}
		if (tree) {
			format_oid(oid_string, regid, regid_length);
			proto_tree_add_text(smux_tree, offset, length,
			    "Registration: %s", oid_string);
		}
		g_free(regid);
		offset += length;

		ret = asn1_octet_string_decode (&asn1, &application, 
		    &application_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree, 
			    "application", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(smux_tree, offset, length,
			    "Application: %.*s", application_length,
			     application);
		}
		g_free(application);
		offset += length;

		ret = asn1_octet_string_decode (&asn1, &password, 
		    &password_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree, 
			    "password", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(smux_tree, offset, length,
			    "Password: %.*s", password_length, password);
		}
		g_free(password);
		offset += length;
		return;
	}
	if (cls == ASN1_APL && con == ASN1_PRI && pdu_type == SMUX_MSG_CLOSE) {
		pdu_type_string = val_to_str(pdu_type, smux_types,
		    "Unknown PDU type %#x");
		if (check_col(fd, COL_INFO))
			col_add_str(fd, COL_INFO, pdu_type_string);
		length = asn1.pointer - start;
		if (tree) {
			proto_tree_add_text(smux_tree, offset, length,
			    "PDU type: %s", pdu_type_string);
		}
		offset += length;
		ret = asn1_uint32_value_decode (&asn1, pdu_length, &cause);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "cause", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(smux_tree, offset, pdu_length,
			    "Cause: %s",
			    val_to_str(cause, smux_close, 
				"Unknown cause %#x"));
		}
		offset += pdu_length;
		return;
	}
	if (cls == ASN1_APL && con == ASN1_CON && pdu_type == SMUX_MSG_RREQ) {
		pdu_type_string = val_to_str(pdu_type, smux_types,
		    "Unknown PDU type %#x");
		if (check_col(fd, COL_INFO))
			col_add_str(fd, COL_INFO, pdu_type_string);
		length = asn1.pointer - start;
		if (tree) {
			proto_tree_add_text(smux_tree, offset, length,
			    "PDU type: %s", pdu_type_string);
		}
		offset += length;
		ret = asn1_oid_decode (&asn1, &regid, &regid_length, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "registration subtree", ret);
			return;
		}
		if (tree) {
			format_oid(oid_string, regid, regid_length);
			proto_tree_add_text(smux_tree, offset, length,
			    "Registration: %s", oid_string);
		}
		g_free(regid);
		offset += length;

		ret = asn1_uint32_decode (&asn1, &priority, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "priority", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(smux_tree, offset, length,
			    "Priority: %d", priority);
		}
		offset += length;

		ret = asn1_uint32_decode (&asn1, &operation, &length);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "operation", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(smux_tree, offset, length,
			    "Operation: %s", 
			    val_to_str(operation, smux_rreq, 
				"Unknown operation %#x"));
		}
		offset += length;
		return;
	}
	if (cls == ASN1_APL && con == ASN1_PRI && pdu_type == SMUX_MSG_RRSP) {
		pdu_type_string = val_to_str(pdu_type, smux_types,
		    "Unknown PDU type %#x");
		if (check_col(fd, COL_INFO))
			col_add_str(fd, COL_INFO, pdu_type_string);
		length = asn1.pointer - start;
		if (tree) {
			proto_tree_add_text(smux_tree, offset, length,
			    "PDU type: %s", pdu_type_string);
		}
		offset += length;
		ret = asn1_uint32_value_decode (&asn1, pdu_length, &priority);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "priority", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(smux_tree, offset, pdu_length,
			    "%s",
			    val_to_str(priority, smux_prio, 
				"Priority: %#x"));
		}
		offset += pdu_length;
		return;
	}
	if (cls == ASN1_APL && con == ASN1_PRI && pdu_type == SMUX_MSG_SOUT) {
		pdu_type_string = val_to_str(pdu_type, smux_types,
		    "Unknown PDU type %#x");
		if (check_col(fd, COL_INFO))
			col_add_str(fd, COL_INFO, pdu_type_string);
		length = asn1.pointer - start;
		if (tree) {
			proto_tree_add_text(smux_tree, offset, length,
			    "PDU type: %s", pdu_type_string);
		}
		offset += length;
		ret = asn1_uint32_value_decode (&asn1, pdu_length, &commit);
		if (ret != ASN1_ERR_NOERROR) {
			dissect_snmp_parse_error(pd, offset, fd, tree,
			    "commit", ret);
			return;
		}
		if (tree) {
			proto_tree_add_text(smux_tree, offset, pdu_length,
			    "%s",
			    val_to_str(commit, smux_sout, 
				"Unknown SOUT Value: %#x"));
		}
		offset += pdu_length;
		return;
	}
	if (cls != ASN1_CTX || con != ASN1_CON) {
		dissect_snmp_parse_error(pd, offset, fd, tree,
		    "PDU type", ASN1_ERR_WRONG_TYPE);
		return;
	}
	dissect_common_pdu(pd, offset, fd, smux_tree, asn1, pdu_type, start);
}

void
dissect_snmp(const u_char *pd, int offset, frame_data *fd, proto_tree *tree) 
{
	dissect_snmp_pdu(pd, offset, fd, tree, "SNMP", proto_snmp, ett_snmp);
}

void
dissect_smux(const u_char *pd, int offset, frame_data *fd, proto_tree *tree) 
{
	dissect_smux_pdu(pd, offset, fd, tree, "SMUX", proto_smux, ett_smux);
}

void
proto_register_snmp(void)
{
        static hf_register_info hf[] = {
		{ &hf_snmpv3_flags,
		{ "SNMPv3 Flags", "snmpv3.flags", FT_UINT8, BASE_HEX, NULL,
		    0x0, "" }},
		{ &hf_snmpv3_flags_auth,
		{ "Authenticated", "snmpv3.flags.auth", FT_BOOLEAN, 8,
		    TFS(&flags_set_truth), TH_AUTH, "" }},
		{ &hf_snmpv3_flags_crypt,
		{ "Encrypted", "snmpv3.flags.crypt", FT_BOOLEAN, 8,
		    TFS(&flags_set_truth), TH_CRYPT, "" }},
		{ &hf_snmpv3_flags_report,
		{ "Reportable", "snmpv3.flags.report", FT_BOOLEAN, 8,
		    TFS(&flags_set_truth), TH_REPORT, "" }},
/*
                { &hf_variable,
                { "Name",           "snmp.abbreviation", TYPE, VALS_POINTER }},
*/
        };
	static gint *ett[] = {
		&ett_snmp,
		&ett_smux,
		&ett_global,
		&ett_flags,
	};

	smiInit("ethereal");
        proto_snmp = proto_register_protocol("Simple Network Management Protocol", "snmp");
        proto_smux = proto_register_protocol("SNMP Multiplex Protocol", "smux");
        proto_register_field_array(proto_snmp, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
}

void
proto_reg_handoff_snmp(void)
{
	dissector_add("udp.port", UDP_PORT_SNMP, dissect_snmp);
	dissector_add("udp.port", UDP_PORT_SNMP_TRAP, dissect_snmp);
	dissector_add("tcp.port", TCP_PORT_SMUX, dissect_smux);
	dissector_add("ethertype", ETHERTYPE_SNMP, dissect_snmp);
}
